/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.session;

import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.SessionScoped;
import javax.faces.component.UIComponent;
import javax.faces.component.UIInput;
import javax.faces.context.FacesContext;

import kieker.analysis.display.annotation.Display;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIDisplayConnector;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.beans.IProjectBean;
import kieker.webgui.beans.application.ProjectsBean;
import kieker.webgui.common.FSManager;
import kieker.webgui.common.Global;
import kieker.webgui.common.exception.NewerProjectException;

import org.primefaces.context.RequestContext;
import org.primefaces.event.TabChangeEvent;

/**
 * This bean contains the project of the current (session) user for the cockpit editor.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
@ManagedBean
@SessionScoped
public class CurrentCockpitEditorBean implements IProjectBean {
	/**
	 * This is the log for errors, exceptions etc.
	 */
	private static final Log LOG = LogFactory.getLog(CurrentCockpitEditorBean.class);
	/**
	 * This is the factory which will be used to create new components for the project.
	 */
	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();
	/**
	 * This is the name of the stored project. It can be used as an identifier within the FS-Manager
	 */
	private String projectName;
	/**
	 * This is the actual model instance. It is the in-memory-model of the current (session) user.
	 */
	private MIProject project;
	/**
	 * This is the time stamp of the moment, the project was loaded or last saved. It can be used to check whether the project has been modified in the meanwhile.
	 */
	private long timeStamp;
	/**
	 * This is the currently selected view.
	 */
	private MIView activeView;

	/**
	 * Creates a new instance of this class.
	 */
	public CurrentCockpitEditorBean() {
		// No code necessary
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation.
	 * 
	 * @param newName
	 *            The name of the project.
	 * @return The name of the page for the analysis view work space.
	 */
	@Override
	public String setProject(final String newName) {
		synchronized (this) {
			// Remember the given parameters
			this.project = ProjectsBean.getInstance().openProject(newName);
			this.projectName = newName;

			if (this.project != null) {
				// Remember the current time! This is important for the later comparison of the time stamps.
				this.resetTimeStamp();
			}
			// Now deliver the correct navigation page
			final String navigationPage;
			if (this.project != null) {
				navigationPage = Global.PAGE_ANALYSIS_VIEW_WORK_SPACE;
			} else {
				navigationPage = "";
			}
			return navigationPage;
		}
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	@Override
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * This method clears the bean. In other words: The stored project is set to null and it will return the page of the project overview for navigation purposes.
	 * 
	 * @return The name of the page of the project overview.
	 */
	@Override
	public String clearProject() {
		synchronized (this) {
			this.projectName = null; // NOPMD
			this.project = null; // NOPMD
			this.activeView = null; // NOPMD
			this.timeStamp = 0;
		}

		return Global.PAGE_PROJECT_OVERVIEW;
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail.
	 * 
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 */
	public void saveProject(final boolean overwriteNewerProject) {
		synchronized (this) {
			try {
				FSManager.getInstance().saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject);
				CurrentCockpitEditorBean.showMessage(FacesMessage.SEVERITY_INFO, "Project saved.");
				// Update the time stamp!
				this.resetTimeStamp();
			} catch (final IOException ex) {
				CurrentCockpitEditorBean.LOG.error("An error occured while saving the projct.", ex);
				CurrentCockpitEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while saving the projct.");
			} catch (final NewerProjectException ex) {
				CurrentCockpitEditorBean.LOG.info("The project has been modified externally in the meanwhile.", ex);
				CurrentCockpitEditorBean.showMessage(FacesMessage.SEVERITY_WARN, "The project has been modified externally in the meanwhile.");
				// Give the user the possibility to force-save the project
				RequestContext.getCurrentInstance().execute("forceSaveDlg.show()");
			}
		}
	}

	/**
	 * This method shows the current user a message by using the growl-component of PrimeFaces.
	 * 
	 * @param severity
	 *            The severity of the message.
	 * @param msg
	 *            The message itself.
	 */
	private static void showMessage(final Severity severity, final String msg) {
		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(severity, "", msg));
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		synchronized (this) {
			this.timeStamp = System.currentTimeMillis();
		}
	}

	/**
	 * This method delivers the available displays of the given plugin.
	 * 
	 * @param clazz
	 *            The plugin whose displays should be delivered.
	 * @return The available displays.
	 */
	public List<Display> getDisplays(final Class<AbstractPlugin> clazz) {
		final List<Display> result = new ArrayList<Display>();

		synchronized (this) {
			if (clazz != null) {
				// Run through all available methods and find the annotated ones
				final Method[] methods = clazz.getMethods();
				for (final Method method : methods) {
					final Display displayAnnot = method.getAnnotation(Display.class);
					// Check whether the annotation is available. If yes - add it to our result list
					if (displayAnnot != null) {
						result.add(displayAnnot);
					}
				}
			}
		}

		return result;
	}

	/**
	 * This method adds a new view to the project.
	 * 
	 * @param viewName
	 *            The name of the new view.
	 */
	public void addView(final String viewName) {
		synchronized (this) {
			if (this.project != null) {
				// Create the view and add it to our project
				final MIView view = this.factory.createView();
				view.setName(viewName);
				view.setDescription("No description available.");
				this.project.getViews().add(view);
			}
		}
	}

	/**
	 * Delivers the currently selected view.
	 * 
	 * @return The currently active view.
	 */
	public MIView getActiveView() {
		synchronized (this) {
			return this.activeView;
		}
	}

	/**
	 * This method sets the active view to a new value.
	 * 
	 * @param view
	 *            The new active view.
	 */
	public void setActiveView(final MIView view) {
		synchronized (this) {
			this.activeView = view;
		}
	}

	/**
	 * This method adds the given display to the currently active view. If no view exists, this method does nothing.
	 * 
	 * @param display
	 *            The display which should be added to the current view.
	 */
	public void addDisplayToView(final MIDisplay display) {
		synchronized (this) {
			if (this.activeView != null) {
				final MIDisplayConnector connector = this.factory.createDisplayConnector();
				connector.setDisplay(display);
				connector.setName(UUID.randomUUID().toString());
				this.activeView.getDisplayConnectors().add(connector);
			}
		}
	}

	/**
	 * This event is used if the currently selected tab changes. It modifies the active view property within this bean.
	 * 
	 * @param event
	 *            The onChange-event.
	 */
	public void onChange(final TabChangeEvent event) {
		if (event.getData() instanceof MIView) {
			this.setActiveView((MIView) event.getData());
		}
	}

	/**
	 * This method checks whether a display connector with the given name exists already.
	 * 
	 * @param name
	 *            The name to be checked.
	 * @return true iff the name exists already.
	 */
	private boolean existsDisplayConnectorName(final String name) {
		synchronized (this) {
			// Make sure a view is selected
			if (this.activeView == null) {
				return false;
			}

			// Run through all display connectors and check the name against the given one
			for (final MIDisplayConnector connector : this.activeView.getDisplayConnectors()) {
				if (connector.getName().equals(name)) {
					return true;
				}
			}

			// The name has not been found
			return false;
		}
	}

	/**
	 * This method is used as a validator for new display connector names.
	 * 
	 * @param context
	 *            The context of the validation.
	 * @param toValidate
	 *            The components which has be validated.
	 * @param value
	 *            The new value.
	 */
	public void validateDisplayConnectorName(final FacesContext context, final UIComponent toValidate, final Object value) {
		synchronized (this) {
			final boolean nameExists = this.existsDisplayConnectorName((String) value);
			((UIInput) toValidate).setValid(!nameExists);
		}
	}

}
