/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.common;

import java.io.IOException;
import java.lang.Thread.State;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import kieker.analysis.AnalysisController;
import kieker.analysis.AnalysisController.AnalysisControllerWithMapping;
import kieker.analysis.AnalysisController.STATE;
import kieker.analysis.display.AbstractDisplay;
import kieker.analysis.display.HtmlText;
import kieker.analysis.display.Image;
import kieker.analysis.display.PlainText;
import kieker.analysis.display.annotation.Display;
import kieker.analysis.exception.AnalysisConfigurationException;
import kieker.analysis.model.analysisMetaModel.MIDisplayConnector;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.exception.AnalysisNotInstantiatedException;
import kieker.webgui.common.exception.AnalysisNotRunningException;
import kieker.webgui.common.exception.ProjectAlreadyStartedException;
import kieker.webgui.common.exception.ProjectStillRunningException;

/**
 * This manager is responsible for the currently used and running instances of {@code AnalysisController}. It supplies methods to check the states of analysis,
 * instantiate, start and to stop them. This is also a singleton instance.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
// TODO How to differ between views and displays with same names?
public final class ACManager { // NOCS (Class Data Abstraction Coupling)
	/**
	 * This is the log for errors, exceptions etc.
	 */
	private static final Log LOG = LogFactory.getLog(ACManager.class);
	/**
	 * This is the maximal time the application will wait for an analysis thread to be terminated.
	 */
	private static final long MAX_THREAD_WAIT_TIME_MS = 1000;
	/**
	 * This is the singleton instance of this class.
	 */
	private static final ACManager INSTANCE = new ACManager();
	/**
	 * This list contains the current analysis controllers and their corresponding threads. Not every project does have a controller, but every project can have
	 * maximal one.
	 */
	private final ConcurrentHashMap<String, Triple<AnalysisController, Thread, UpdateDisplaysThread>> analysisController = new ConcurrentHashMap<String,
			Triple<AnalysisController, Thread, UpdateDisplaysThread>>();

	/**
	 * Creates a new instance of this class.
	 */
	private ACManager() {
		// No code necessary.
	}

	/**
	 * Delivers the singleton instance of this class.
	 * 
	 * @return The only instance of {@link FSManager}.
	 */
	public static ACManager getInstance() {
		return ACManager.INSTANCE;
	}

	/**
	 * This method tries to stop a running analysis.
	 * 
	 * @param project
	 *            The project to be stopped.
	 * @throws AnalysisNotRunningException
	 *             If the analysis is not running or has not yet been initialized.
	 * @throws InterruptedException
	 *             If the current thread has somehow been interrupted while waiting for the existing analysis thread.
	 */
	public void stopAnalysisController(final String project) throws AnalysisNotRunningException, InterruptedException {
		final Triple<AnalysisController, Thread, UpdateDisplaysThread> currController = this.analysisController.get(project);
		try {
			synchronized (currController) {
				// Try to stop the analysis
				currController.getFst().terminate();
				currController.getThd().terminate();
				currController.getSnd().join(ACManager.MAX_THREAD_WAIT_TIME_MS);
				currController.getThd().join(ACManager.MAX_THREAD_WAIT_TIME_MS);
			}
		} catch (final NullPointerException ex) { // NOPMD
			ACManager.LOG.info("Analysis is not running.", ex);
			throw new AnalysisNotRunningException(); // NOPMD
		}
	}

	/**
	 * This method tries to start the analysis for the given project. It has already to be initialized.
	 * 
	 * @param project
	 *            The project to be started.
	 * @throws ProjectAlreadyStartedException
	 *             If the project is already running.
	 * @throws AnalysisNotInstantiatedException
	 *             If the analysis has not yet been initialized.
	 */
	public void startAnalysisController(final String project) throws ProjectAlreadyStartedException, AnalysisNotInstantiatedException {
		final Triple<AnalysisController, Thread, UpdateDisplaysThread> currController = this.analysisController.get(project);
		// Is there an analysis after all?
		try {
			synchronized (currController) {
				// Check whether this analysis has already been started - or is already dead.
				if (currController.getSnd().getState() != State.NEW) {
					throw new ProjectAlreadyStartedException();
				}
				// Otherwise start the analysis
				currController.getSnd().start();
				currController.getThd().start();
			}
		} catch (final NullPointerException ex) { // NOPMD (Generic exception)
			ACManager.LOG.info("Analysis is not instantiated.", ex);
			throw new AnalysisNotInstantiatedException(); // NOPMD
		}
	}

	/**
	 * This method cleans the analysis controller of the given project. In other words: The potential running instance is stopped and removed. After a call to this
	 * method, the project can be restarted. Potential exceptions will be ignored
	 * 
	 * @param project
	 *            The name of the project to be cleaned.
	 */
	public void cleanAnalysisController(final String project) {
		final Triple<AnalysisController, Thread, UpdateDisplaysThread> currController = this.analysisController.get(project);
		// Is there an analysis after all?
		try {
			synchronized (currController) {
				this.stopAnalysisController(project);
				this.analysisController.remove(project);
			}
		} catch (final NullPointerException ex) { // NOPMD
			ACManager.LOG.info("Analysis is not instantiated.", ex);
		} catch (final AnalysisNotRunningException ex) {
			ACManager.LOG.info("Analysis is not running.", ex);
		} catch (final InterruptedException ex) {
			ACManager.LOG.info("Unexpected interrupted exception.", ex);
		}

	}

	/**
	 * This method tries to instantiate the analysis controller of the given project using the data available on the file system.
	 * 
	 * @param project
	 *            The project whose analysis controller should be instantiated.
	 * @throws NullPointerException
	 *             If an error occurred while loading the project.
	 * @throws AnalysisConfigurationException
	 *             If the given configuration is somehow invalid.
	 * @throws IOException
	 *             If an error occurred while loading the project.
	 * @throws ProjectStillRunningException
	 *             If there exists already an analysis controller and is still running.
	 */
	public void instantiateAnalysisController(final String project) throws NullPointerException, AnalysisConfigurationException, IOException,
			ProjectStillRunningException {
		final Triple<AnalysisController, Thread, UpdateDisplaysThread> newController = new Triple<AnalysisController, Thread, UpdateDisplaysThread>();
		synchronized (newController) {
			// Add the potential new project atomically
			final Triple<AnalysisController, Thread, UpdateDisplaysThread> currController = this.analysisController.putIfAbsent(project, newController);
			// We can just add a new controller if there wasn't an old one.
			if (currController != null) {
				throw new ProjectStillRunningException("The project with the name '" + project + "' is still running.");
			}
			final MIProject modelProject = FSManager.getInstance().openProject(project);
			final ClassLoader classLoader = FSManager.getInstance().getClassLoader(project);
			final AnalysisControllerWithMapping controller = AnalysisController.createAnalysisController(modelProject, classLoader);

			// Create the necessary threads for the analysis
			final Thread runningThread = new ControllerRunningThread(controller);
			// runningThread.start();

			final UpdateDisplaysThread displayThread = new UpdateDisplaysThread(controller.getPluginMap(), modelProject);
			// Put everything into our container
			newController.setFst(controller.getController());
			newController.setSnd(runningThread);
			newController.setThd(displayThread);
		}
	}

	/**
	 * This method can be used to deliver the state of the analysis controller of the given project as a human readable string.
	 * 
	 * @param project
	 *            The project whose state should be delivered.
	 * @return The current state of the corresponding {@link AnalysisController} if it exists, '' otherwise.
	 */
	public String getAnalysisControllerStateString(final String project) {
		final STATE controllerState = this.getAnalysisControllerState(project);
		final String controllerStateString;

		if (controllerState == null) {
			controllerStateString = "";
		} else {
			controllerStateString = controllerState.toString();
		}

		return controllerStateString;
	}

	/**
	 * Delivers the display object for the given display in the given view in the given project.
	 * 
	 * @param project
	 *            The name of the project.
	 * @param viewName
	 *            The name of the view.
	 * @param displayName
	 *            The name of the display.
	 * @return The display object if it exists, null otherwise.
	 */
	public AbstractDisplay getDisplay(final String project, final String viewName, final String displayName) {
		try {
			return this.analysisController.get(project).getThd().getDisplay(viewName, displayName);
		} catch (final NullPointerException ex) { // NOPMD
			ACManager.LOG.info("Invalid project, view or display.", ex);
			return null;
		}
	}

	/**
	 * This method can be used to deliver the state of the analysis controller of the given project as a human readable string.
	 * 
	 * @param project
	 *            The project whose state should be delivered.
	 * @return The current state of the corresponding {@link AnalysisController} if it exists, null otherwise.
	 */
	public STATE getAnalysisControllerState(final String project) {
		// If the given project is null, we return a null-state.
		if (project == null) {
			return null;
		}
		final Triple<AnalysisController, Thread, UpdateDisplaysThread> controller = this.analysisController.get(project);

		final STATE controllerState;
		if (controller == null) {
			controllerState = null;
		} else {
			synchronized (controller) {
				if (controller.getFst() == null) {
					controllerState = null;
				} else {
					controllerState = controller.getFst().getState();
				}
			}
		}

		return controllerState;
	}

	/**
	 * This method delivers the current displays of the given project. They can be used to show the stored information but should <b>not</b> be modified.
	 * 
	 * @param projectName
	 *            The name of the project whose displays should be delivered.
	 * @return A list of pairs containing the displays and the name of them.
	 */
	public List<Pair<String, AbstractDisplay>> getCurrentDisplays(final String projectName) {
		final List<Pair<String, AbstractDisplay>> result = new ArrayList<Pair<String, AbstractDisplay>>();

		return result;
	}

	/**
	 * This helper thread is responsible for running an instance of {@link AnalysisControllerWithMapping}.
	 * 
	 * @author Nils Christian Ehmke
	 * @version 1.0
	 */
	private static class ControllerRunningThread extends Thread {
		/**
		 * The controller managed by this thread.
		 */
		private final AnalysisControllerWithMapping controller;

		/**
		 * Default constructor.
		 * 
		 * @param controller
		 *            The controller to be managed by this thread.
		 */
		public ControllerRunningThread(final AnalysisControllerWithMapping controller) {
			this.controller = controller;
		}

		@Override
		public void run() {
			try {
				this.controller.getController().run();
			} catch (final IllegalStateException ex) {
				ACManager.LOG.error("An error occured while starting the analysis.", ex);
			} catch (final AnalysisConfigurationException ex) {
				ACManager.LOG.error("An error occured while starting the analysis.", ex);
			}
		}
	}

	/**
	 * This helper thread is used to update the available displays of the given analysis at regular intervals. <b>Important:</b> If the interrupt-method of the
	 * thread is being called, it will be terminated.
	 * 
	 * @author Nils Christian Ehmke
	 * @version 1.0
	 */
	private static class UpdateDisplaysThread extends Thread {

		/**
		 * This is the time the thread waits between the updates.
		 */
		private static final long SLEEP_TIME_MS = 2 * 1000;
		/**
		 * The map between the instances of {@link MIPlugin} and {@link AbstractPlugin} within the running analysis instance.
		 */
		private final Map<MIPlugin, AbstractPlugin> myPluginMap;
		/**
		 * The model instance.
		 */
		private final MIProject myProject;
		/**
		 * The map containing all available display objects.
		 */
		private final Map<String, Map<String, AbstractDisplay>> displayObjects = new ConcurrentHashMap<String, Map<String, AbstractDisplay>>(); // NOPMD
		/**
		 * This map contains the mapping to get the methods to be called.
		 */
		private final Map<AbstractDisplay, Method> methodMap = new ConcurrentHashMap<AbstractDisplay, Method>(); // NOPMD (Use of concurrent hash map)
		/**
		 * The field determining whether the thread has been terminated or not. // (USeConcurrentHashMap)
		 */
		private volatile boolean terminated = false;

		/**
		 * Creates a new instance of this class using the given parameters.
		 * 
		 * @param pluginMap
		 *            The map between the instances of {@link MIPlugin} and {@link AbstractPlugin} within the running analysis instance.
		 * @param modelProject
		 *            The model instance.
		 */
		public UpdateDisplaysThread(final Map<MIPlugin, AbstractPlugin> pluginMap, final MIProject modelProject) {
			this.myPluginMap = pluginMap;
			this.myProject = modelProject;

			// Initialize the hashmaps and the necessary objects
			for (final MIView view : this.myProject.getViews()) {
				final Map<String, AbstractDisplay> viewMap = new ConcurrentHashMap<String, AbstractDisplay>(); // NOPMD (Use of concurrent hash map)
				this.displayObjects.put(view.getName(), viewMap);
				for (final MIDisplayConnector displayConnector : view.getDisplayConnectors()) {
					final Method displayMethod = UpdateDisplaysThread.getDisplayMethod(this.myPluginMap.get(displayConnector.getDisplay().getParent()).getClass(),
							displayConnector.getName());

					// Make sure that the method really exists and that is has the correct parameters
					if ((displayMethod != null) && (displayMethod.getParameterTypes().length == 1)) {
						final Class<?> parameterType = displayMethod.getParameterTypes()[0];
						final AbstractDisplay displayObject;
						// Get the correct type
						if (Image.class.isAssignableFrom(parameterType)) {
							displayObject = new Image();
						} else {
							if (PlainText.class.isAssignableFrom(parameterType)) {
								displayObject = new PlainText();
							} else {
								if (HtmlText.class.isAssignableFrom(parameterType)) {
									displayObject = new HtmlText();
								} else {
									displayObject = null;
								}
							}
						}
						if (displayObject != null) {
							viewMap.put(displayConnector.getName(), displayObject);
							this.methodMap.put(displayObject, displayMethod);
						}
					}
				}
			}
		}

		/**
		 * Delivers the correct method to be invoked for an update of the display.
		 * 
		 * @param clazz
		 *            The class of the plugin.
		 * @param displayName
		 *            The name within the display-annotation.
		 * @return The method instance.
		 */
		private static Method getDisplayMethod(final Class<? extends AbstractPlugin> clazz, final String displayName) {
			final Method[] methods = clazz.getMethods();
			for (final Method method : methods) {
				final Display displayAnnot = method.getAnnotation(Display.class);
				if ((displayAnnot != null) && displayAnnot.name().equals(displayName)) {
					// We found the correct method
					return method;
				}
			}
			return null;
		}

		/**
		 * Terminates this thread.
		 */
		public void terminate() {
			this.terminated = true;
		}

		/**
		 * Delivers the display object for the given view and the given display.
		 * 
		 * @param viewName
		 *            The name of the view.
		 * @param displayName
		 *            The name of the display in the view.
		 * @return The object for the given display if it exists, null otherwise.
		 */
		public AbstractDisplay getDisplay(final String viewName, final String displayName) {
			return this.displayObjects.get(viewName).get(displayName);
		}

		@Override
		public void run() {
			// Run until we have been interrupted
			while (!this.terminated) {
				for (final MIView view : this.myProject.getViews()) {
					final Map<String, AbstractDisplay> viewMap = this.displayObjects.get(view.getName()); // NOPMD (Use of concurrent hash map)
					for (final MIDisplayConnector displayConnector : view.getDisplayConnectors()) {
						final AbstractDisplay displayObject = viewMap.get(displayConnector.getName());
						final AbstractPlugin pluginObject = this.myPluginMap.get(displayConnector.getDisplay().getParent());
						// Update the display object
						try {
							this.methodMap.get(displayObject).invoke(pluginObject, displayObject);
						} catch (final IllegalAccessException ex) {
							ACManager.LOG.error("An error occured while updating the display.", ex);
						} catch (final IllegalArgumentException ex) {
							ACManager.LOG.error("An error occured while updating the display.", ex);
						} catch (final InvocationTargetException ex) {
							ACManager.LOG.error("An error occured while updating the display.", ex);
						} catch (final NullPointerException ex) { // NOPMD
							ACManager.LOG.error("An error occured while updating the display.", ex);
						}
					}
				}

				// Wait a little bit.
				try {
					Thread.sleep(UpdateDisplaysThread.SLEEP_TIME_MS);
				} catch (final InterruptedException ex) {
					// We have been interrupted. Exit the thread
					return;
				}
			}
		}
	}
}
